-------------------------------------------------------------------------------
--  Copyright (c) 2016, 2017 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------
--  06/07/17  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : Fast Inter-board Data Interface                                                          
--  PCB           : any
--  Entity Name   : PS2_KEYBOARD_INF
--  File Name     : PS2_KEYBOARD_INF.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  PS2_KEYBOARD_INF.vhd is uses the following components:
--    none
--
--  PS2_KEYBOARD_INF.vhd is uses the following packages:
--    Package        none
--
-------------------------------------------------------------------------------
--  PS2_KEYBOARD_INF.vhd
--
--  PS2_KEYBOARD_INF detects PS2 Keyboard scan codes. The design follows the functional
--  description in the Digilent Genesys Reference Manual for the PS2 interface.
--
--  This component handles most of the commonly used keyboard key scan codes and
--  indicated if the message represents a make (key press) or break (key release). It 
--  also indicates if the key is an extended key such as the right ATL key to the right
--  of the space bar.  
--
--  The interface:
--    CLK_FREQUENCY_MHZ A generic input specifying the frequency of clk in MHz
--    clk               The clock input. Must be have the frrequency equal to CLK_FREQUENCY_MHZ
--    areset            An asynchronous reset input asserted high
--    ps2_clk           The unconditioned clock signal from the PS2 keyboard connector
--    ps2_data          The unconditioned data signal from the PS2 keybard connector
--    keycode           An output represeting the key scan code
--    key_make          Asserted high if the key detection is a key press
--    key_break         Asserted high if the key press is a key release
--    key_extended      Asserted high if the key has an extended code association
--    key_valid         Asserted high if no parity errors were detected
--    newkey_s          Asserted high for 1 clk period. When high indicated that keycode,
--                      key_make, key_break, key_extended and key_valid outputs are valid. 
--  General usage:
--    When the newkey_s strobe is asserted high the user should latch the other outputs. As
--    the PS2 keyboard sends codes for both key presses and key releases the user must decide
--    which newkey_s event to ignore. Some keys such as the ALT, SHIFT, CAPS LOCK might have
--    make-break events that don't occur consecutively. For example the captial H key is
--    represented by a SHIFT-key make event followed by an H-key make event followed by an
--    H-key break event and perhaps followed by a SHIFT-key break event.  
--
--    INACTIVE_DETECT_PERIOD_US represents the elapsed time between 11-bit scan codes and
--    is used to determing if these 11-bit scan codes are related to the same keyboard key
--    event. Currently, this value is a constant set to 200 us. The actual value depends on
--    how a particular keyboard is configured. 
-- 
--  General implementation:
--
--    The way that I implemented this design is to have a 33-bit shift register that can
--    accomodate up to 3 11-bit scan codes for any keyboard event. This limits its usage to
--    keys with one scan code plus a break scan code plus an extended scan code. The design 
--    assumes that scan code come in the following order <extended><break>[key code] where
--    <codes> are optional. The design take into account the fact that extended key make
--    messages are [extended][key code]. It does not handle other keys which have more than
--    one key code. 
-- 
-------------------------------------------------------------------------------
--  Known unresolved Issues:
--
-------------------------------------------------------------------------------
--  Limitations
--    
--    - This version does not support writing commands to the keyboard
--    - This version does not detect keyboard error messages
--    - This version does not support keys with multiple scan codes ( beyond the
--      break and extended codes )
--
------------------------------------------------------------------------------
--  ToDo
--
--    INACTIVE_DETECT_PERIOD_US should be a generic instead of a constant.
--
-------------------------------------------------------------------------------
--  NOTES:
--
-------------------------------------------------------------------------------
--  CAUTIONS!!!
--
--    This component assumes that there is no keyboard bounce; that is there will be only
--    one make and one break event per key press.
-- 
-------------------------------------------------------------------------------

Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;
  use ieee.math_real.all; 
  
   
entity PS2_KEYBOARD_INF is
  generic(
          CLK_FREQUENCY_MHZ    : natural := 50);
  port(
        clk           : in   std_logic;
        areset        : in   std_logic;
        ps2_clk       : in   std_logic;
        ps2_data      : in   std_logic;
        keycode       : out  std_logic_vector(7 downto 0);
        key_make      : out  std_logic;
        key_break     : out  std_logic;
        key_extended  : out  std_logic;
        key_valid     : out  std_logic;
        newkey_s      : out  std_logic
  );
end entity;

architecture rtl of PS2_KEYBOARD_INF is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

--type sm_states is (IDLE, 


-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

constant  INACTIVE_DETECT_PERIOD_US  : natural := 200; 
constant  INACTIVE_DETECT_CLOCKS     : natural := integer(round(real(CLK_FREQUENCY_MHZ)*real(INACTIVE_DETECT_PERIOD_US)));
constant  TIMER_WIDTH                : natural := integer(ceil(log2(real(INACTIVE_DETECT_CLOCKS))));
constant  MAKE_KEY_BIT_COUNT_M1             : natural := 10;
constant  BREAK_KEY_BIT_COUNT_M1            : natural := 21;
constant  BREAK_EXTENDED_KEY_BIT_COUNT_M1   : natural := 32;
constant  BREAK_CODE                  : std_logic_vector(7 downto 0) := X"F0";
constant  EXTENDED_CODE               : std_logic_vector(7 downto 0) := X"E0";


-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  ps2_clk_reg       : std_logic_vector(2 downto 0);
signal  ps2_sample_s      : std_logic;
signal  data_sr           : std_logic_vector(10 downto 0);
signal  bitcount          : std_logic_vector(6 downto 0);
signal  timer             : std_logic_vector(TIMER_WIDTH-1 downto 0);
signal  ps2_is_idle       : std_logic;
signal  ps2_is_idle_r     : std_logic;
signal  report_s          : std_logic;
signal  parity_error      : std_logic;
signal  code_reg_b        : std_logic_vector(7 downto 0); 
signal  code_reg_e        : std_logic_vector(7 downto 0); 
signal  dreg0             : std_logic_vector(10 downto 0);
signal  dreg1             : std_logic_vector(10 downto 0);

begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------
                    
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  ps2_sample_s    <= '1' when ps2_clk_reg = "100" else '0';  -- sample data on falling edge
  ps2_is_idle     <= '1' when timer = conv_std_logic_vector(INACTIVE_DETECT_CLOCKS,TIMER_WIDTH) else '0';
  report_s        <= '1' when ps2_is_idle_r = '0' and ps2_is_idle = '1' else '0';
  
  --verify that parity, start, and stop bits are all correct
  parity_error <= NOT (NOT data_sr(0) 
                       AND data_sr(10) 
                       AND (data_sr(9) XOR data_sr(8) XOR data_sr(7) XOR data_sr(6) XOR data_sr(5) 
                                       XOR data_sr(4) XOR data_sr(3) XOR data_sr(2) XOR data_sr(1))
                      );  
                     
                      
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------

  -- PS2 clock synchronization
  process(clk,areset)
  begin
    if areset = '1' then
      ps2_clk_reg   <= (others => '1');
    elsif rising_edge(clk) then
      ps2_clk_reg   <= ps2_clk_reg((ps2_clk_reg'left)-1 downto 0) & ps2_clk;    
    end if;       
  end process;
    
  -- keyboard scan code shift register
  -- data arrives start, D0, D1....D7, parity, stop
  -- since data is guaranteed valid at least 5 us before and after the sample
  -- period there is no need to synchronize the keyboard data line.
  process(clk,areset)
  begin
    if areset = '1' then
      data_sr   <= (others => '0');
    elsif rising_edge(clk) then
      if ps2_sample_s = '1' then
        data_sr   <= ps2_data & data_sr(data_sr'left downto 1);
      end if;      
    end if;       
  end process;
  
  -- a timer to detect a quiescent state
  process(clk,areset)
  begin
    if areset = '1' then
      timer   <= (others => '0');
    elsif rising_edge(clk) then
      if ps2_clk_reg = "000" then
        timer   <= (others => '0');
      elsif timer <  conv_std_logic_vector(INACTIVE_DETECT_CLOCKS,TIMER_WIDTH) then
        timer   <= timer + 1;
      end if;      
    end if;       
  end process;
  
  -- count bits to determine if a message represents
  -- a make, break or extended code
  -- As the first falling edge of the PS2 clock happens before ps2_is_idle is asserted
  -- we miss counting the start bit; so bitcount is low by 1 
  process(clk,areset)
  begin
    if areset = '1' then
      bitcount   <= (others => '0');
    elsif rising_edge(clk) then
      if ps2_is_idle = '1' then
        bitcount   <= (others => '0');
      elsif ps2_sample_s = '1' and ps2_is_idle = '0' then
        bitcount   <= bitcount + 1;
      end if;      
    end if;       
  end process;
  
  -- create a 1 clock delayed version of ps2_is_idle so that we can create a 1 clock
  -- wide strobe to indicate that the data and output flags are valid
  process(clk,areset)
  begin
    if areset = '1' then
      ps2_is_idle_r   <= '0';
    elsif rising_edge(clk) then
      ps2_is_idle_r   <= ps2_is_idle;    
    end if;       
  end process;
  
  
  
  -- save first code in message
  process(clk,areset)
  begin
    if areset = '1' then
      dreg0   <= (others => '0');
    elsif rising_edge(clk) then
      if bitcount = conv_std_logic_vector(MAKE_KEY_BIT_COUNT_M1,6) then
        dreg0   <= data_sr;
      end if;      
    end if;       
  end process;

  
  -- save second code in message
  process(clk,areset)
  begin
    if areset = '1' then
      dreg1   <= (others => '0');
    elsif rising_edge(clk) then
      if bitcount = conv_std_logic_vector(BREAK_KEY_BIT_COUNT_M1,6) then
        dreg1   <= data_sr;
      end if;      
    end if;       
  end process;
                                                     
  -- save the extended character code
  process(clk,areset)
  begin
    if areset = '1' then
      code_reg_e   <= (others => '0');
    elsif rising_edge(clk) then
      if bitcount = conv_std_logic_vector(BREAK_KEY_BIT_COUNT_M1,6) then
        code_reg_e   <= dreg0(8 downto 1);
      elsif bitcount = conv_std_logic_vector(BREAK_EXTENDED_KEY_BIT_COUNT_M1,6) then
        code_reg_e   <= dreg0(8 downto 1);
      elsif bitcount = "0000000" then  
        code_reg_e   <= (others => '0');
      end if;      
    end if;       
  end process;
        
  -- save the break code
  process(clk,areset)
  begin
    if areset = '1' then
      code_reg_b   <= (others => '0');
    elsif rising_edge(clk) then
      if bitcount = conv_std_logic_vector(BREAK_KEY_BIT_COUNT_M1,6) then
        code_reg_b   <= dreg0(8 downto 1);
      elsif bitcount = conv_std_logic_vector(BREAK_EXTENDED_KEY_BIT_COUNT_M1,6) then
        code_reg_b   <= dreg1(8 downto 1);
      elsif bitcount = "0000000" then  
        code_reg_b   <= (others => '0');
      end if;      
    end if;       
  end process;
  
    
-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
                     
  keycode       <= data_sr(8 downto 1); 
  key_make      <= '1' when bitcount = conv_std_logic_vector(MAKE_KEY_BIT_COUNT_M1,6) else
                   '1' when bitcount = conv_std_logic_vector(BREAK_KEY_BIT_COUNT_M1,6) and code_reg_e = EXTENDED_CODE else
                   '0'; 
  key_break     <= '1' when code_reg_b = BREAK_CODE else '0';
  key_extended  <= '1' when code_reg_e = EXTENDED_CODE else '0'; 
  key_valid     <= '1' when report_s = '1' and parity_error = '0' else '0';
  newkey_s      <= report_s;
 
    
end rtl;
