-------------------------------------------------------------------------------
--  Copyright (c) 2017 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in its original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------
--  07/03/17  W.C.  Created                                     
--  07/04/17  W.C.  Corrected an error 
--  07/05/17  W.C.  Fixed missing areset in process sensitivity list                                    
--
-------------------------------------------------------------------------------
--  Project       : any                                                          
--  PCB           : any
--  Device        : any
--  Entity Name   : DEBOUNCE
--  File Name     : DEBOUNCE.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  DEBOUNCE.vhd is uses the following components:
--
--  DEBOUNCE.vhd is uses the following packages:
--    Package        none
--
-------------------------------------------------------------------------------
--  DEBOUNCE.vhd
--
--    DEBOUNCE conditions the input signal bouncy_in so that it can only change
--    state after being stable for at least the number of microseconds set by the
--    generic MIN_SWITCHING_PERIOD_US. 
--
-------------------------------------------------------------------------------
--  Limitations
--
------------------------------------------------------------------------------
--  ToDo
--
-------------------------------------------------------------------------------
--  NOTES:
--
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;
  use ieee.math_real.all;
  
entity DEBOUNCE is
  generic(
    CLK_RATE_MHZ              : natural := 100;
    MIN_SWITCHING_PERIOD_US   : natural := 1000);   
  port(
        clk                   : in    std_logic; -- a clock whose frequency = CLK_RATE_MHZ * 1000000
        areset                : in    std_logic; -- an asynchronous reset
        bouncy_in             : in    std_logic; -- an unconditioned signal
        unbouncy_out          : out   std_logic  -- a conditioned version of bouncy_in    
  );
end entity;

architecture rtl of DEBOUNCE is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

  constant SWITCH_THRESHOLD_COUNT   : natural := integer(round(real(CLK_RATE_MHZ)*real(MIN_SWITCHING_PERIOD_US)));
  constant PERIOD_COUNTER_WIDTH     : natural := integer(ceil(log2(real(SWITCH_THRESHOLD_COUNT))));
  constant PERIOD_COUNT_ZERO        : std_logic_vector(PERIOD_COUNTER_WIDTH-1 downto 0) := (others => '0');


-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal period_count         : std_logic_vector(PERIOD_COUNTER_WIDTH-1 downto 0);
signal bouncy_in_r          : std_logic;
signal bouncy_in_r1         : std_logic;

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------

  process(clk,areset)
  begin
    if areset = '1' then
      bouncy_in_r   <= bouncy_in;
      bouncy_in_r1  <= bouncy_in;
    elsif rising_edge(clk) then
      bouncy_in_r   <= bouncy_in;
      if bouncy_in_r = bouncy_in and period_count = SWITCH_THRESHOLD_COUNT-1 then
        bouncy_in_r1  <= bouncy_in_r;
      end if;  
    end if;    
  end process;

   
  process(clk,areset)
  begin
    if areset = '1' then
      period_count   <= (others => '0');
    elsif rising_edge(clk) then
      if period_count < SWITCH_THRESHOLD_COUNT then
        period_count   <= period_count + 1;
      elsif bouncy_in_r1 /= bouncy_in_r then
        period_count   <= (others => '0');
      end if;
    end if;    
  end process;

-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------

  unbouncy_out  <= bouncy_in_r1;  

end rtl;
