-------------------------------------------------------------------------------
--  Copyright (c) 2017 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in its original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------
--  06/26/17  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : IO_Expander                                                          
--  PCB           : Digilent Nexys Video
--  Device        : XC7A200T-1SBG484C
--  Entity Name   : NV
--  File Name     : NV.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  NV_Test.vhd is uses the following components:
--    DEBOUNCE  DEBOUNCE.vhd
--
--  NV_Test.vhd is uses the following packages:
--    Package        none
--
-------------------------------------------------------------------------------
--  NV_Test.vhd
--
--
--    This entity is a "proof-of-concept" demonstrating that Ethernet PHYs can be used
--    be used to connect two different FPGA boards and allow each access to the other's
--    resources. Specifically, the purpose of this demonstration is to demonstrate one
--    way to increase the sparse IO capabilities of the Nexys Video board by allowing
--    the Nexys Video to access the Spartan 3A board's IO.
--
--    As this project is only meant to demonstrate "proof-of-concept" minimal functionality
--    is provided. The objective to to relect some inputs on the other board and to provide
--    a few pins that will allow measuring latency on an oscilloscope.
--
--    When the Spartan 3A board detects a change in the state of 4 user switches or input 
--    from the PS2 keyboard it sends a packet to the Nexys Video through the attached
--    Ethernet cable. The 4-bit switch state and low 4 bits of the key scan code are reflected
--    on both Nexys Video LEDs and JA PMOD. The low 4 bits of the key scan code are also
--    sent to J20. When the Nexys Video detects a change in the 8 user switches or JB pins 
--    [10:7] it sends a packet to the 3A board where the switches are relected in the LEDs
--    and the JB pins are reflected on J18.
--
--    Both boards have a similar design. There are separate clock domains for sampling inputs
--    and sending packets and for receiving packets and updating outputs. A counter creates a 
--    1 us sample period for detecting input changes and possibly generating Tx packets. The
--    Ethernet receiver is available to update outputs at any time.  
--
--    The RTL8211E supplies the Rx clock and uses an GMII 4-bit SDR data interface once it has 
--    auto-negotiated a data rate with the Spartan 3A PHY.
--
--    The data packets sent by both boards consists of 5 bytes: 2-byte preamble, 2 byte payload,
--    and one byte terminator. The board receiving packets will not update outputs unless the
--    packet is validated by having the correct preamble and terminator bytes. No error other
--    checking is performed.
--
-------------------------------------------------------------------------------
--  Limitations
--
------------------------------------------------------------------------------
--  ToDo
--
-------------------------------------------------------------------------------
--  NOTES:
--
--    - Before trying to send packets:
--      - configure both the Nexys Video and Spartan 3A boards
--      - connect a good quality short Ethernet CAT 5E or better cable to the Ethernet jacks on
--        both boards
--      - Reset the Nexys Video PHY using RED CPU_RESET button so that it can negotiate a
--        100 Mbps data rate with the LAN8700 PHY on the 3A board
--      - Once the Nexys Video RTL8211E LINK and USER LEDs are lit you are ready to go 
--
-------------------------------------------------------------------------------

Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;
  use ieee.math_real.all;

library UNISIM;
  use UNISIM.vcomponents.all;
  
entity NV is
  Port( 
        sysclk            : in    std_logic; -- global external 100 MHz clock        
        cpu_resetn        : in    std_logic; -- normally '1', '0' when Reset button pressed 
        
        -- General User IO
        btnc              : in    std_logic; -- Vadj HR Bank 16 normally '0', '1' when pressed
--        btnd      : in    std_logic; -- Vadj HR Bank 16 normally '0', '1' when pressed
--        btnl      : in    std_logic; -- Vadj HR Bank 16 normally '0', '1' when pressed
--        btnr      : in    std_logic; -- Vadj HR Bank 16 normally '0', '1' when pressed
--        btnu      : in    std_logic; -- Vadj HR Bank 16 normally '0', '1' when pressed
        sw                : in    std_logic_vector(7 downto 0); -- Vadj HR Bank 15 or 16
        led               : out   std_logic_vector(7 downto 0); -- VCC2V5 HR bank 13 
              
        -- PMOD connectors
        JA                : out   std_logic_vector(7 downto 0);
        JB                : inout std_logic_vector(7 downto 0);
        -- Ethernet PHY interface
        eth_mdc           : out   std_logic;  
        eth_mdio          : inout std_logic;  
        eth_reset_n       : out   std_logic;  
        eth_rxclk         : in    std_logic;  
        eth_rxctl         : in    std_logic;  
        eth_rxd           : in    std_logic_vector(3 downto 0); 
        eth_txclk         : out   std_logic;  
        eth_txctl         : out   std_logic;  
        eth_txd           : out   std_logic_vector(3 downto 0)  
  );
end entity;

architecture rtl of NV is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

component DEBOUNCE
  generic(
    CLK_RATE_MHZ              : natural := 100;
    MIN_SWITCHING_PERIOD_US   : natural := 1000);   
  port(
        clk                   : in    std_logic;
        areset                : in    std_logic; 
        bouncy_in             : in    std_logic;
        unbouncy_out          : out   std_logic      
  );
end component;

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

type etx_sm_states is (t_idle, t_sample, t_load, t_send);
type erx_sm_states is (r_idle, r_div, r_checkv, r_wrdata);
                                                   
-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

  -- interface packet definition
  -- <preamble (2) ><payload (2) ><terminator (1) >
  -- NV payload bytes: <3A_LEDS><3A_J18_IO>
  -- 3A payload bytes: <NV_LEDS><NV_JA>
  constant PREAMBLE_SYMBOLS         : std_logic_vector(15 downto 0) := X"555D";
  constant TERMINATOR_SYMBOL        : std_logic_vector(7 downto 0) := X"AA";
  constant NUM_PREAMBLE_BYTES       : natural := 2;  
  constant NUM_PAYLOAD_BYTES        : natural := 2;  
  constant NUM_TERMINATOR_BYTES     : natural := 1;
  constant NUM_PACKET_BYTES         : natural := NUM_PREAMBLE_BYTES + NUM_PAYLOAD_BYTES + NUM_TERMINATOR_BYTES;
  constant SHIFT_REGISTER_WIDTH     : natural := 8*NUM_PACKET_BYTES;   
  constant UPDATE_RATE_KHZ          : natural := 1000; -- 1 us input sample period 
  constant TX_CLOCK_RATE_KHZ        : natural := 25000;
  constant SAMPLE_STROBE_COUNT      : natural := integer(round(real(TX_CLOCK_RATE_KHZ)/real(UPDATE_RATE_KHZ)));
  constant PERIOD_COUNTER_WIDTH     : natural := integer(ceil(log2(real(SAMPLE_STROBE_COUNT))));
   
  constant TEST_HALF_PERIOD_COUNT   : std_logic_vector(19 downto 0) := X"7A120"; -- 500000 = 100 Hz @ 100 MHz
-------------------------------------------------------------------------------
--  Function Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  current_tstate        : etx_sm_states;
signal  current_rstate        : erx_sm_states;


signal  clk100          : std_logic;
signal  clk25           : std_logic;
signal  clk25p90        : std_logic;
signal  clkfb           : std_logic;
signal  pll_locked      : std_logic;
signal  areset          : std_logic;
signal  eth_tx_sr       : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  eth_rx_sr       : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  send_data_stb   : std_logic;
signal  send_data       : std_logic;
signal  byte_count      : std_logic_vector(5 downto 0);
signal  input_event     : std_logic;
signal  debounced_sw    : std_logic_vector(7 downto 0);
signal  sw_reg          : std_logic_vector(7 downto 0);
signal  jb_h_reg        : std_logic_vector(3 downto 0);
signal  ja_reg          : std_logic_vector(7 downto 0);
signal  led_reg         : std_logic_vector(7 downto 0); 
signal  eth_tx_payload  : std_logic_vector(8*NUM_PAYLOAD_BYTES-1 downto 0); 
signal  eth_rx_preamble : std_logic_vector(8*NUM_PREAMBLE_BYTES-1 downto 0);
signal  eth_rx_payload  : std_logic_vector(8*NUM_PAYLOAD_BYTES-1 downto 0);
signal  eth_rx_term     : std_logic_vector(8*NUM_TERMINATOR_BYTES-1 downto 0);
signal  valid_rx_packet : std_logic;
signal  period_count    : std_logic_vector(8*PERIOD_COUNTER_WIDTH-1 downto 0);
signal  jb0_test_count  : std_logic_vector(19 downto 0);
-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

  -- PLLE2_BASE: Base Phase Locked Loop (PLL)
  -- 7 Series
  -- Xilinx HDL Libraries Guide, version 13.4
  PLLE2_BASE_inst : PLLE2_BASE
  generic map (
    BANDWIDTH => "OPTIMIZED", -- OPTIMIZED, HIGH, LOW
    CLKFBOUT_MULT   => 10, -- Multiply value for all CLKOUT, (2-64)
    CLKFBOUT_PHASE  => 0.0, -- Phase offset in degrees of CLKFB, (-360.000-360.000).
    CLKIN1_PERIOD   => 10.0, -- Input clock period in ns to ps resolution (i.e. 33.333 is 30 MHz).
    -- CLKOUT0_DIVIDE - CLKOUT5_DIVIDE: Divide amount for each CLKOUT (1-128)
    CLKOUT0_DIVIDE  => 10,
    CLKOUT1_DIVIDE  => 40, -- 8
    CLKOUT2_DIVIDE  => 40, -- 8
    CLKOUT3_DIVIDE  => 10,
    CLKOUT4_DIVIDE  => 10,
    CLKOUT5_DIVIDE  => 10,
    -- CLKOUT0_DUTY_CYCLE - CLKOUT5_DUTY_CYCLE: Duty cycle for each CLKOUT (0.001-0.999).
    CLKOUT0_DUTY_CYCLE  => 0.5,
    CLKOUT1_DUTY_CYCLE  => 0.5,
    CLKOUT2_DUTY_CYCLE  => 0.5,
    CLKOUT3_DUTY_CYCLE  => 0.5,
    CLKOUT4_DUTY_CYCLE  => 0.5,
    CLKOUT5_DUTY_CYCLE  => 0.5,
    -- CLKOUT0_PHASE - CLKOUT5_PHASE: Phase offset for each CLKOUT (-360.000-360.000).
    CLKOUT0_PHASE       => 0.0,
    CLKOUT1_PHASE       => 0.0,
    CLKOUT2_PHASE       => 90.0,
    CLKOUT3_PHASE       => 0.0,
    CLKOUT4_PHASE       => 0.0,
    CLKOUT5_PHASE       => 0.0,
    DIVCLK_DIVIDE       => 1, -- Master division value, (1-56)
    REF_JITTER1         => 0.0, -- Reference input jitter in UI, (0.000-0.999).
    STARTUP_WAIT        => "FALSE" -- Delay DONE until PLL Locks, ("TRUE"/"FALSE")
  )
  port map (
    -- Clock Outputs: 1-bit (each) output: User configurable clock outputs
    CLKOUT0     => clk100,      -- 100 MHz
    CLKOUT1     => clk25,      -- 125 MHz
    CLKOUT2     => clk25p90,   -- 125 MHz phase shifted 90 degrees
    CLKOUT3     => open,
    CLKOUT4     => open,
    CLKOUT5     => open,
    -- Feedback Clocks: 1-bit (each) output: Clock feedback ports
    CLKFBOUT    => clkfb, -- 1-bit output: Feedback clock
    -- Status Port: 1-bit (each) output: PLL status ports
    LOCKED      => pll_locked, -- 1-bit output: LOCK
    -- Clock Input: 1-bit (each) input: Clock input
    CLKIN1      => sysclk, -- 1-bit input: Input clock
    -- Control Ports: 1-bit (each) input: PLL control ports
    PWRDWN      => '0', -- 1-bit input: Power-down
    RST         => '0', -- 1-bit input: Reset
    -- Feedback Clocks: 1-bit (each) input: Clock feedback ports
    CLKFBIN     => clkfb -- 1-bit input: Feedback clock
  );

  GEN_debounce : for i in 7 downto 0 generate
    DEBOUNCE_inst : DEBOUNCE
    generic map(
      CLK_RATE_MHZ              => 25,
      MIN_SWITCHING_PERIOD_US   => 2000) -- 2 milliseconds  
    port map(
          clk             => clk25,
          areset          => areset, 
          bouncy_in       => sw(i),
          unbouncy_out    => debounced_sw(i)     
          );                 
  end generate;
    
           
--------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  areset          <= not cpu_resetn;
  
  input_event     <= '0' when sw_reg = debounced_sw and jb_h_reg = JB(7 downto 4) else '1';
                     
  send_data_stb   <= '1' when current_tstate = t_load else '0';
  valid_rx_packet <= '1' when (eth_rx_preamble = PREAMBLE_SYMBOLS) and (eth_rx_term = TERMINATOR_SYMBOL) else '0';
  
  eth_tx_payload  <= debounced_sw & X"0" & JB(7 downto 4);
  
  eth_rx_preamble <= eth_rx_sr(eth_rx_sr'left downto eth_rx_sr'left-(8*NUM_PREAMBLE_BYTES-1));        -- [39:24]
  eth_rx_payload  <= eth_rx_sr(eth_rx_sr'left-(8*NUM_PREAMBLE_BYTES) downto 8*NUM_TERMINATOR_BYTES);  -- [23:8]
  eth_rx_term     <= eth_rx_sr(8*NUM_TERMINATOR_BYTES-1 downto 0);                                    -- [7:0]
  
  send_data       <= '1' when byte_count > "000000" else
                     '1' when current_tstate = t_load else
                     '0'; 
  
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------

  -- a 100 Hz test signal for JB(0) to drive JB(4)
  process(clk100, areset)
  begin
    if areset = '1' then
      jb0_test_count  <= (others => '0');
    elsif rising_edge(clk100) then
      if jb0_test_count < TEST_HALF_PERIOD_COUNT(TEST_HALF_PERIOD_COUNT'left -1 downto 0) & '0' then -- 2*TEST_HALF_PERIOD_COUNT18
        jb0_test_count  <= jb0_test_count + 1;
      else  
        jb0_test_count  <= (others => '0');
      end if;
    end if;  
  end process;

  -- the sample period counter
  process(clk25, areset)
  begin
    if areset = '1' then
      period_count  <= (others => '0');
    elsif rising_edge(clk25) then
      if period_count < SAMPLE_STROBE_COUNT then
        period_count  <= period_count + 1;
      else   
        period_count  <= (others => '0');
      end if;
    end if;  
  end process;

  -- track Tx bytes ( nibbles for the 3A )
  process(clk25, areset)
  begin
    if areset = '1' then
      byte_count  <= (others => '0');
    elsif rising_edge(clk25) then
      if send_data_stb = '1' then
        byte_count  <= conv_std_logic_vector(2*NUM_PACKET_BYTES,6); -- count nibbles
      elsif byte_count > "000000" then
        byte_count  <= byte_count - 1;
      end if;
    end if;  
  end process;
  
  -- Ethernet TX data shift register
  process(clk25, areset)
  begin
    if areset = '1' then
      eth_tx_sr  <= PREAMBLE_SYMBOLS & eth_tx_payload & TERMINATOR_SYMBOL;
    elsif rising_edge(clk25) then
      if send_data_stb = '1' then
        eth_tx_sr  <= PREAMBLE_SYMBOLS & eth_tx_payload & TERMINATOR_SYMBOL;
      elsif send_data = '1' then   
        eth_tx_sr  <= eth_tx_sr(eth_tx_sr'left -4 downto 0) & eth_tx_sr(eth_tx_sr'left downto eth_tx_sr'left -3);
      end if;
    end if;  
  end process;
  
  -- Ethernet RX data shift register
  process(eth_rxclk, areset)
  begin
    if areset = '1' then
      eth_rx_sr  <= (others => '0');
    elsif rising_edge(eth_rxclk) then
      if current_rstate = r_idle then
        eth_rx_sr  <= (others => '0');
      elsif eth_rxctl = '1' then   
        eth_rx_sr  <= eth_rx_sr(eth_rx_sr'left -4 downto 0) & eth_rxd;
      end if;
    end if;  
  end process;
  
  
  -- a state machine to control the Ethernet transmitter
-- type etx_sm_states is (t_idle, t_sample, t_load, t_send);
  ETX_SM : process (clk25, areset)
  begin
    if (areset = '1') then
      sw_reg          <= (others => '0');
      jb_h_reg        <= (others => '0');
      current_tstate  <= t_idle;
    elsif rising_edge(clk25) then
      case current_tstate is
        -- wait until period counter matches sample count
        when t_idle =>
          if period_count = SAMPLE_STROBE_COUNT then  
            current_tstate   <= t_sample;
          else
            current_tstate  <= t_idle;
          end if;
          
        -- check for changes          
        when t_sample =>
          sw_reg    <= debounced_sw;
          jb_h_reg  <= JB(7 downto 4);
          if input_event = '1' then
            current_tstate  <= t_load;
          else   
            current_tstate  <= t_idle;           
          end if;  
         
        -- send a packet if anything has changed          
        when t_load =>
          current_tstate  <= t_send;
                   
        when t_send =>
          current_tstate  <= t_idle;
          
        when others =>
          current_tstate  <= t_idle;
        
      end case;
    end if;      
  end process ETX_SM; 



  -- a state machine to control the Ethernet transmitter
-- type erx_sm_states is (r_idle, r_div, r_checkv, r_wrdata);
  ERX_SM : process (eth_rxclk, areset)
  begin
    if (areset = '1') then
      ja_reg    <= (others => '0');
      led_reg   <= (others => '0');  
      current_rstate  <= r_idle;
    elsif rising_edge(eth_rxclk) then
      case current_rstate is                                
        when r_idle =>                                      
          if eth_rxctl = '1' then  
            current_rstate   <= r_div;
          else
            current_rstate  <= r_idle;
          end if;
            
        -- wait until all eth_rx data is in       
        when r_div =>
          if eth_rxctl = '0' then
            current_rstate  <= r_checkv;
          else   
            current_rstate  <= r_div;           
          end if;  
                 
        -- check that the preamble and termination bytes are correct  
        when r_checkv =>
          if valid_rx_packet = '1' then
            current_rstate  <= r_wrdata;
          else   
            current_rstate  <= r_idle;           
          end if;  
             
        -- only write if the packet is valid      
        when r_wrdata =>
          led_reg   <= eth_rx_payload(15 downto 8); -- need to use constants 
          ja_reg    <= eth_rx_payload(7 downto 0); -- need to use constants
          current_rstate  <= r_idle;
          
        when others =>
          current_rstate  <= r_idle;
        
      end case;
    end if;      
  end process ERX_SM; 
  
-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
  
  eth_mdc       <= '0';  
  eth_mdio      <= 'Z'; 
  eth_reset_n   <= cpu_resetn;
  eth_txclk     <= clk25p90;
  eth_txd       <= eth_tx_sr(eth_tx_sr'left downto eth_tx_sr'left -3);     
  eth_txctl     <= send_data;
  
  JA              <= ja_reg;
  JB(0)           <= '1' when (jb0_test_count > TEST_HALF_PERIOD_COUNT) and (debounced_sw(7) = '1') else '0';
  JB(3 downto 1)  <= (others => 'Z'); -- unused outputs
  JB(7 downto 4)  <= (others => 'Z'); -- always inputs
   
  led           <= led_reg;
end rtl;
