-------------------------------------------------------------------------------
--  Copyright (c) 2017 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in its original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------
--  06/26/17  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : Ethernet IO Expander                                                          
--  PCB           : Digilent Spartan 3A Starter Kit Board (Rev C)
--  Device        : XC3S700A-FGG484C -4
--  Entity Name   : Spartan3A
--  File Name     : Spartan3A.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  Spartan3A_Test.vhd is uses the following components:
--    PS2_KEYBOARD_INF  PS2_KEYBOARD_INF.vhd
--    DEBOUNCE          DEBOUNCE.vhd
--
--  Spartan3A_Test.vhd is uses the following packages:
--    Package        none
--
-------------------------------------------------------------------------------
--  Spartan3A_Test.vhd
--
--    This entity is a "proof-of-concept" demonstrating that Ethernet PHYs can be used
--    be used to connect two different FPGA boards and allow each access to the other's
--    resources. Specifically, the purpose of this demonstration is to demonstrate one
--    way to increase the sparse IO capabilities of the Nexys Video board by allowing
--    the Nexys Video to access the Spartan 3A board's IO.
--
--    As this project is only meant to demonstrate "proof-of-concept" minimal functionality
--    is provided. The objective to to relect some inputs on the other board and to provide
--    a few pins that will allow measuring latency on an oscilloscope.
--
--    When the Spartan 3A board detects a change in the state of 4 user switches or input 
--    from the PS2 keyboard it sends a packet to the Nexys Video through the attached
--    Ethernet cable. The 4-bit switch state and low 4 bits of the key scan code are reflected
--    on both Nexys Video LEDs and JA PMOD. The low 4 bits of the key scan code are also
--    sent to J20. When the Nexys Video detects a change in the 8 user switches or JB pins 
--    [10:7] it sends a packet to the 3A board where the switches are relected in the LEDs
--    and the JB pins are reflected on J18.
--
--    Both boards have a similar design. There are separate clock domains for sampling inputs
--    and sending packets and for receiving packets and updating outputs. A counter creates a 
--    1 us sample period for detecting input changes and possibly generating Tx packets. The
--    Ethernet receiver is available to update outputs at any time.  
--
--    The LAN8700 Ethernet PHY supplies both the Rx and Tx clocks and has a 4-bit SDR MII 
--    data interface. 
--
--    The data packets sent by both boards consists of 5 bytes: 2-byte preamble, 2 byte payload,
--    and one byte terminator. The board receiving packets will not update outputs unless the
--    packet is validated by having the correct preamble and terminator bytes. No error other
--    checking is performed.
--
-------------------------------------------------------------------------------
--  Limitations
--
------------------------------------------------------------------------------
--  ToDo
--
-------------------------------------------------------------------------------
--  NOTES:
--
--    - Before trying to send packets:
--      - configure both the Nexys Video and Spartan 3A boards
--      - connect a good quality short Ethernet CAT 5E or better cable to the connectors on
--        both boards
--      - Reset the Nexys Video PHY using RED CPU_RESET button so that it can negotiate a
--        100 Mbps data rate with the LAN8700 PHY on the 3A board 
--      - Once the Nexys Video RTL8211E LINK and USER LEDs are lit you are ready to go 
--
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;
  use ieee.math_real.all;
  
entity Spartan3A is
  port(
       CLK_50MHZ     : in    std_logic;
       
       LED           : out   std_logic_vector(7 downto 0);
       
       -- user buttons
       BTN_NORTH     : in    std_logic;   -- '1' when pressed
       
       -- user switches
       SW            : in    std_logic_vector(3 downto 0);
       
       -- Ethernet Interface
       E_MDC         : out   std_logic;      --Ethernet PHY
       E_MDIO        : inout std_logic;      --management data in/out
       E_COL         : in    std_logic;
       E_CRS         : in    std_logic;
       E_NRST        : out   std_logic;
       E_NINT        : in    std_logic;
       E_RX_CLK      : in    std_logic;      --receive clock
       E_RX_ERR      : in    std_logic;      --receive error
       E_RX_DV       : in    std_logic;      --data valid
       E_RXD         : in    std_logic_vector(3 downto 0);
       E_TX_CLK      : in    std_logic;      --transmit clock
       E_TX_EN       : out   std_logic;      --data valid
       E_TX_ER       : out   std_logic;      --transmit error
       E_TXD         : out   std_logic_vector(3 downto 0);
       
       -- PS2 Keyboard interface
       PS2_CLK1      :in     std_logic;
       PS2_DATA1     :in     std_logic;
       
       -- J18
       J18_IO        : out   std_logic_vector(3 downto 0);
       J19_IO        : out   std_logic_vector(3 downto 0);
       J20_IO        : out   std_logic_vector(3 downto 0)
        
  );
end entity;

architecture rtl of Spartan3A is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

component PS2_KEYBOARD_INF
  generic(
          CLK_FREQUENCY_MHZ    : natural := 50);
  port(
        clk           : in   std_logic;
        areset        : in   std_logic;
        ps2_clk       : in   std_logic;
        ps2_data      : in   std_logic;
        keycode       : out  std_logic_vector(7 downto 0);
        key_make      : out  std_logic;
        key_break     : out  std_logic;
        key_extended  : out  std_logic;
        key_valid     : out  std_logic;
        newkey_s      : out  std_logic
  );
end component;

component DEBOUNCE
  generic(
    CLK_RATE_MHZ              : natural := 100;
    MIN_SWITCHING_PERIOD_US   : natural := 1000);   
  port(
        clk                   : in    std_logic;
        areset                : in    std_logic; 
        bouncy_in             : in    std_logic;
        unbouncy_out          : out   std_logic      
  );
end component;

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

type etx_sm_states is (t_idle, t_sample, t_load, t_send);
type erx_sm_states is (r_idle, r_div, r_checkv, r_wrdata);

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

  -- interface packet definition
  -- <preamble (2) ><payload (2) ><terminator (1) >
  -- NV payload bytes: <3A_LEDS><3A_J18_IO>
  -- 3A payload bytes: <NV_LEDS><NV_JA>
  constant PREAMBLE_SYMBOLS         : std_logic_vector(15 downto 0) := X"555D";
  constant TERMINATOR_SYMBOL        : std_logic_vector(7 downto 0) := X"AA";
  constant NUM_PREAMBLE_BYTES       : natural := 2;  
  constant NUM_PAYLOAD_BYTES        : natural := 2;  
  constant NUM_TERMINATOR_BYTES     : natural := 1;
  constant NUM_PACKET_BYTES         : natural := NUM_PREAMBLE_BYTES + NUM_PAYLOAD_BYTES + NUM_TERMINATOR_BYTES;
  constant SHIFT_REGISTER_WIDTH     : natural := 8*NUM_PACKET_BYTES;    
  constant UPDATE_RATE_KHZ          : natural := 1000; -- 1 us input sample period 
  constant TX_CLOCK_RATE_KHZ        : natural := 25000;
  constant SAMPLE_STROBE_COUNT      : natural := integer(round(real(TX_CLOCK_RATE_KHZ)/real(UPDATE_RATE_KHZ)));
  constant PERIOD_COUNTER_WIDTH     : natural := integer(ceil(log2(real(SAMPLE_STROBE_COUNT))));

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  current_tstate        : etx_sm_states;
signal  current_rstate        : erx_sm_states;

signal  areset          : std_logic;
signal  eth_tx_sr       : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  eth_rx_sr       : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  btn_lo_cnt      : std_logic_vector(25 downto 0);
signal  send_data_stb   : std_logic;
signal  send_data       : std_logic;
signal  byte_count      : std_logic_vector(5 downto 0);
signal  irxd            : std_logic_vector(3 downto 0);
signal  irxdiv          : std_logic;
signal  irxdiv_r        : std_logic;
signal  irxdiv_stb      : std_logic;
signal  itxd            : std_logic;
signal  input_event     : std_logic;
signal  debounced_sw    : std_logic_vector(3 downto 0);
signal  sw_reg          : std_logic_vector(3 downto 0);
signal  j20_reg         : std_logic_vector(3 downto 0);
signal  j18_reg         : std_logic_vector(3 downto 0);
signal  led_reg         : std_logic_vector(7 downto 0); 
signal  eth_tx_payload  : std_logic_vector(8*NUM_PAYLOAD_BYTES-1 downto 0); 
signal  eth_rx_preamble : std_logic_vector(8*NUM_PREAMBLE_BYTES-1 downto 0);
signal  eth_rx_payload  : std_logic_vector(8*NUM_PAYLOAD_BYTES-1 downto 0);
signal  eth_rx_term     : std_logic_vector(8*NUM_TERMINATOR_BYTES-1 downto 0);
signal  valid_rx_packet : std_logic;
signal  period_count    : std_logic_vector(8*PERIOD_COUNTER_WIDTH-1 downto 0);
signal  keycode_update  : std_logic;
signal  keymake         : std_logic;
signal  keybreak        : std_logic;
signal  keyextended     : std_logic;
signal  keyvalid        : std_logic;
signal  keycode         : std_logic_vector(7 downto 0);
signal  keycode_r       : std_logic_vector(7 downto 0);
signal  keycode_r1      : std_logic_vector(7 downto 0);

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------
    
  KB0 : PS2_KEYBOARD_INF
  generic map(
          CLK_FREQUENCY_MHZ => TX_CLOCK_RATE_KHZ/1000)
  port map(
        clk           => E_TX_CLK,
        areset        => areset,
        ps2_clk       => PS2_CLK1,
        ps2_data      => PS2_DATA1,
        keycode       => keycode,
        key_make      => keymake,
        key_break     => keybreak,
        key_extended  => keyextended,
        key_valid     => keyvalid,
        newkey_s      => keycode_update);

  GEN_debounce : for i in 3 downto 0 generate
    DEBOUNCE_inst : DEBOUNCE
    generic map(
      CLK_RATE_MHZ              => 25,
      MIN_SWITCHING_PERIOD_US   => 2000) -- 2 milliseconds  
    port map(
          clk             => E_TX_CLK,
          areset          => areset, 
          bouncy_in       => SW(i),
          unbouncy_out    => debounced_sw(i)     
          );                 
  end generate;

-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  areset  <= BTN_NORTH;
  irxd    <= E_RXD;
  irxdiv  <= E_RX_DV;
  
  
  input_event     <= '0' when sw_reg = debounced_sw and j20_reg = keycode_r(3 downto 0) else '1';
  send_data_stb   <= '1' when current_tstate = t_load else '0';
  valid_rx_packet <= '1' when (eth_rx_preamble = PREAMBLE_SYMBOLS) and (eth_rx_term = TERMINATOR_SYMBOL) else '0';
  
  -- send j20_reg to upper 4 NV LEDs and JA(8:5)
  -- send SW to lower 4 NV LEDs and JA(4:1)
  eth_tx_payload  <= j20_reg & debounced_sw & j20_reg & j18_reg;
  
  eth_rx_preamble <= eth_rx_sr(eth_rx_sr'left downto eth_rx_sr'left-(8*NUM_PREAMBLE_BYTES-1));        -- [39:24]
  eth_rx_payload  <= eth_rx_sr(eth_rx_sr'left-(8*NUM_PREAMBLE_BYTES) downto 8*NUM_TERMINATOR_BYTES);  -- [23:8]
  eth_rx_term     <= eth_rx_sr(8*NUM_TERMINATOR_BYTES-1 downto 0);                                    -- [7:0]
  
            
  
  send_data       <= '1' when byte_count > "000000" else
                     '1' when current_tstate = t_load else
                     '0'; 
                     
  irxdiv_stb  <= '1' when current_rstate = r_checkv else '0';
  
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------
  
  -- capture make key scan codes
  process(E_TX_CLK, areset)
  begin
    if areset = '1' then
      keycode_r  <= (others => '0');
    elsif falling_edge(E_TX_CLK) then
      if keymake = '1' and keyvalid = '1' and keycode_update = '1' then
        keycode_r  <= keycode;  
      end if;
    end if;  
  end process;

  -- the sample period counter
  process(E_TX_CLK, areset)
  begin
    if areset = '1' then
      period_count  <= (others => '0');
    elsif falling_edge(E_TX_CLK) then
      if period_count < SAMPLE_STROBE_COUNT then
        period_count  <= period_count + 1;
      else   
        period_count  <= (others => '0');
      end if;
    end if;  
  end process;

  -- track Tx bytes ( nibbles for the 3A )
  process(E_TX_CLK, areset)
  begin
    if areset = '1' then
      byte_count  <= (others => '0');
    elsif falling_edge(E_TX_CLK) then
      if send_data_stb = '1' then
        byte_count  <= conv_std_logic_vector(2*NUM_PACKET_BYTES,6); -- count nibbles
      elsif byte_count > "000000" then
        byte_count  <= byte_count - 1;
      end if;
    end if;  
  end process;
  
  -- Ethernet TX data shift register
  process(E_TX_CLK, areset)
  begin
    if areset = '1' then
      eth_tx_sr  <= PREAMBLE_SYMBOLS & eth_tx_payload & TERMINATOR_SYMBOL;
    elsif falling_edge(E_TX_CLK) then
      if send_data_stb = '1' then
        eth_tx_sr  <= PREAMBLE_SYMBOLS & eth_tx_payload & TERMINATOR_SYMBOL;
      elsif send_data = '1' then   
        eth_tx_sr  <= eth_tx_sr(eth_tx_sr'left -4 downto 0) & eth_tx_sr(eth_tx_sr'left downto eth_tx_sr'left -3);
      end if;
    end if;  
  end process;
  
  -- Ethernet RX data shift register
  process(E_RX_CLK, areset)
  begin
    if areset = '1' then
      eth_rx_sr  <= (others => '0');
    elsif rising_edge(E_RX_CLK) then
      if current_rstate = r_idle then
        eth_rx_sr  <= (others => '0');
      elsif irxdiv = '1' then   
        eth_rx_sr  <= eth_rx_sr(eth_rx_sr'left -4 downto 0) & irxd;
      end if;
    end if;  
  end process;

  -- a state machine to control the Ethernet transmitter
-- type etx_sm_states is (t_idle, t_sample, t_load, t_send);
  ETX_SM : process (E_TX_CLK, areset)
  begin
    if (areset = '1') then
      sw_reg          <= (others => '0');
      j20_reg         <= (others => '0');
      current_tstate  <= t_idle;
    elsif falling_edge(E_TX_CLK) then
      case current_tstate is
        -- wait until period counter matches sample count
        when t_idle =>
          if period_count = SAMPLE_STROBE_COUNT then  
            current_tstate   <= t_sample;
          else
            current_tstate  <= t_idle;
          end if;
          
        -- check for changes          
        when t_sample =>
          sw_reg   <= debounced_sw;
          j20_reg  <= keycode_r(3 downto 0);
          if input_event = '1' then
            current_tstate  <= t_load;
          else   
            current_tstate  <= t_idle;           
          end if;  
         
        -- send a packet if anything has changed          
        when t_load =>
          current_tstate  <= t_send;
                   
        when t_send =>
          current_tstate  <= t_idle;
          
        when others =>
          current_tstate  <= t_idle;
        
      end case;
    end if;      
  end process ETX_SM; 



  -- a state machine to control the Ethernet transmitter
-- type erx_sm_states is (r_idle, r_div, r_checkv, r_wrdata);
  ERX_SM : process (E_RX_CLK, areset)
  begin
    if (areset = '1') then
      j18_reg   <= (others => '0');
      led_reg   <= (others => '0');  
      current_rstate  <= r_idle;
    elsif rising_edge(E_RX_CLK) then
      case current_rstate is                                
        when r_idle =>                                      
          if irxdiv = '1' then  
            current_rstate   <= r_div;
          else
            current_rstate  <= r_idle;
          end if;
            
        -- wait until all eth_rx data is in       
        when r_div =>
          if irxdiv = '0' then
            current_rstate  <= r_checkv;
          else   
            current_rstate  <= r_div;           
          end if;  
                 
        -- check that the preamble and termination bytes are correct  
        when r_checkv =>
          if valid_rx_packet = '1' then
            current_rstate  <= r_wrdata;
          else   
            current_rstate  <= r_idle;           
          end if;  
             
        -- only write if the packet is valid      
        when r_wrdata =>
          led_reg   <= eth_rx_payload(15 downto 8); -- need to use constants 
          j18_reg   <= eth_rx_payload(3 downto 0); -- need to use constants
          current_rstate  <= r_idle;
          
        when others =>
          current_rstate  <= r_idle;
        
      end case;
    end if;      
  end process ERX_SM; 

-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
  
  -- loopback perhaps
  E_TXD   <= eth_tx_sr(eth_tx_sr'left downto eth_tx_sr'left -3);
  E_TX_EN <= send_data;
  E_TX_ER <= '0';
  
  E_MDC   <= '1';
  E_MDIO  <= 'Z';
  E_NRST  <= not BTN_NORTH;
  
  J18_IO        <= j18_reg;   
  J20_IO        <= j20_reg;
  J19_IO        <= (others => 'Z');
  LED     <= led_reg;

end rtl;
